// This file is part of the program FRYSK.
//
// Copyright 2005, 2007, 2008, Red Hat Inc.
//
// FRYSK is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//
// FRYSK is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with FRYSK; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
// 
// In addition, as a special exception, Red Hat, Inc. gives You the
// additional right to link the code of FRYSK with code not covered
// under the GNU General Public License ("Non-GPL Code") and to
// distribute linked combinations including the two, subject to the
// limitations in this paragraph. Non-GPL Code permitted under this
// exception must only link to the code of FRYSK through those well
// defined interfaces identified in the file named EXCEPTION found in
// the source code files (the "Approved Interfaces"). The files of
// Non-GPL Code may instantiate templates or use macros or inline
// functions from the Approved Interfaces without causing the
// resulting work to be covered by the GNU General Public
// License. Only Red Hat, Inc. may make changes or additions to the
// list of Approved Interfaces. You must obey the GNU General Public
// License in all respects for all of the FRYSK code and other code
// used in conjunction with FRYSK except the Non-GPL Code covered by
// this exception. If you modify this file, you may extend this
// exception to your version of the file, but you are not obligated to
// do so. If you do not wish to provide this exception without
// modification, you must delete this exception statement from your
// version and license this file solely under the GPL without
// exception.

package frysk.sys;

import java.util.Map;
import java.util.WeakHashMap;

/**
 * Host signal handling.
 *
 * Need to use native calls to find out the current host's signal
 * values.
 */
public class Signal implements Comparable {
    private final int sig;
    private final String name;
    private Signal(int sig, String name) {
	this.sig = sig;
	this.name = name;
    }
    private Signal(int sig) {
        this.sig = sig;
        this.name = "SIG" + sig;
    }

    /**
     * Dump the signal to a string.  The format is "NAME(VALUE)".
     */
    public String toString() {
	return name + "(" + this.intValue() + ")";
    }

    /**
     * Hash on the Signal's intValue.
     */
    public int hashCode() {
	return intValue();
    }

    /**
     * Return the signal's numeric value as an integer.
     */
    public int intValue() {
        return sig;
    }

    /**
     * Comparison operations; use intValue() so that sub-classing can
     * provide a scratch int value.
     */
    public boolean equals(Object o) {
        return ((o instanceof Signal)
                && ((Signal) o).intValue() == this.intValue());
    }
    public int compareTo(Object o) {
	Signal rhs = (Signal)o; // exception ok
	return this.intValue() - rhs.intValue();
    }

    /**
     * Deliver SIG to process PID.
     */
    public void kill(ProcessIdentifier pid) {
	kill(pid.intValue(), this.intValue(), this.name);
    }
    private static native void kill(int pid, int signum, String name);
    /**
     * Deliver SIGNAL to task (or thread) LWP.
     */
    public final void tkill(ProcessIdentifier lwp) {
	tkill(lwp.intValue(), this.intValue(), this.name);
    }
    private static native void tkill(int lwp, int signum, String name);

    /**
     * Momentarialy sets the signal handler for Sig to SIGIGN so that
     * any pending signals are discarded.
     */
    public final void drain() {
	drain(sig);
    }
    private static native void drain(int signum);

    /**
     * Return the known <b>host</b> signals as a SignalSet.
     */
    public static SignalSet getHostSignalSet() {
        return new SignalSet(signals);
    }

    /**
     * Given an integer, return the corresponding <b>host</b> signal.
     * If the signal is unknown, make one up.
     *
     * This is method is <b>package-private</b>.  If you need to convert
     * a target signal to a Signal object, see frysk.proc.Signal.
     */
    static Signal valueOf(int signum) {
	if (signum < 0) {
	    throw new NullPointerException("invalid signal: " + signum);
        } else if (signum < signals.length) {
            Signal signal = signals[signum];
            if (signal != null)
                return signal;
            // Fall through to bottom which will make up a signal.
        }
        // Unknown signal, fudge one up; beter than throwing
        // an exception.
        synchronized (unknownSignals) {
            scratchSignal.signum = signum;
            Signal signal = (Signal)unknownSignals.get(scratchSignal);
            if (signal != null)
                return signal;
            signal = new Signal(signum);
            unknownSignals.put(signal, signal);
            return signal;
        }
    }

    /**
     * A scratch value for searching the signal table.
     */
    private static class ScratchSignal extends Signal {
        ScratchSignal() {
           super(0, "<scratch signal>");
       }
        int signum;
        public int intValue() {
           return signum;
        }
    }
    private static final Map unknownSignals = new WeakHashMap();
    private static final ScratchSignal scratchSignal = new ScratchSignal();

    /**
     * Tables for looking up known signals; weak so that scratch
     * signals do not accumulate.
     */
    private static native int nsig();
    private static native int rtMin();
    private static native int rtMax();
    // Note: nsig()
    private static final Signal[] signals = new Signal[nsig()];
    /**
     * Create a signal, also enter the signal into the signals table
     * for later retrieval.
     */
    private static Signal signalFactory(int sig, String name) {
        if (sig < 0)
            return null;
        Signal signal = new Signal(sig, name);
        if (signals[sig] == null)
            signals[sig] = signal;
	return signal;
    }

    public static final Signal NONE = signalFactory(0, "SIGNONE");
    private static native int hup();
    public static final Signal HUP = signalFactory(hup(), "SIGHUP");
    private static native int int_();
    public static final Signal INT = signalFactory(int_(), "SIGINT");
    private static native int quit();
    public static final Signal QUIT = signalFactory(quit(), "SIGQUIT");
    private static native int ill();
    public static final Signal ILL = signalFactory(ill(), "SIGILL");
    private static native int abrt();
    public static final Signal ABRT = signalFactory(abrt(), "SIGABRT");
    private static native int fpe();
    public static final Signal FPE = signalFactory(fpe(), "SIGFPE");
    private static native int kill();
    public static final Signal KILL = signalFactory(kill(), "SIGKILL");
    private static native int segv();
    public static final Signal SEGV = signalFactory(segv(), "SIGSEGV");
    private static native int pipe();
    public static final Signal PIPE = signalFactory(pipe(), "SIGPIPE");
    private static native int alrm();
    public static final Signal ALRM = signalFactory(alrm(), "SIGALRM");
    private static native int term();
    public static final Signal TERM = signalFactory(term(), "SIGTERM");
    private static native int usr1();
    public static final Signal USR1 = signalFactory(usr1(), "SIGUSR1");
    private static native int usr2();
    public static final Signal USR2 = signalFactory(usr2(), "SIGUSR2");
    private static native int chld();
    public static final Signal CHLD = signalFactory(chld(), "SIGCHLD");
    private static native int cont();
    public static final Signal CONT = signalFactory(cont(), "SIGCONT");
    private static native int stop();
    public static final Signal STOP = signalFactory(stop(), "SIGSTOP");
    private static native int tstp();
    public static final Signal TSTP = signalFactory(tstp(), "SIGTSTP");
    private static native int ttin();
    public static final Signal TTIN = signalFactory(ttin(), "SIGTTIN");
    private static native int ttou();
    public static final Signal TTOU = signalFactory(ttou(), "SIGTTOU");
    private static native int bus();
    public static final Signal BUS = signalFactory(bus(), "SIGBUS");
    private static native int prof();
    public static final Signal PROF = signalFactory(prof(), "SIGPROF");
    private static native int sys();
    public static final Signal SYS = signalFactory(sys(), "SIGSYS");
    private static native int trap();
    public static final Signal TRAP = signalFactory(trap(), "SIGTRAP");
    private static native int urg();
    public static final Signal URG = signalFactory(urg(), "SIGURG");
    private static native int vtalrm();
    public static final Signal VTALRM = signalFactory(vtalrm(), "SIGVTALRM");
    private static native int xcpu();
    public static final Signal XCPU = signalFactory(xcpu(), "SIGXCPU");
    private static native int xfsz();
    public static final Signal XFSZ = signalFactory(xfsz(), "SIGXFSZ");
    private static native int emt();
    public static final Signal EMT = signalFactory(emt(), "SIGEMT");
    private static native int stkflt();
    public static final Signal STKFLT = signalFactory(stkflt(), "SIGSTKFLT");
    private static native int io();
    public static final Signal IO = signalFactory(io(), "SIGIO");
    private static native int pwr();
    public static final Signal PWR = signalFactory(pwr(), "SIGPWR");
    private static native int lost();
    public static final Signal LOST = signalFactory(lost(), "SIGLOST");
    private static native int winch();
    public static final Signal WINCH = signalFactory(winch(), "SIGWINCH");
    private static native int unused();
    public static final Signal UNUSED = signalFactory(unused(), "SIGUNUSED");
    private static native int poll();
    public static final Signal POLL = signalFactory(poll(), "SIGPOLL");
    private static native int iot();
    public static final Signal IOT = signalFactory(iot(), "SIGIOT");
    private static native int info();
    public static final Signal INFO = signalFactory(info(), "SIGINFO");
    private static native int cld();
    public static final Signal CLD = signalFactory(cld(), "SIGCLD");
    static {
        // Populate the real-time signal table.  XXX while signal(7)
        // seems to imply rtMax() is valid (cannot "exceed"), the code
        // suggests otherwise.
        for (int sig = rtMin(); sig < rtMax(); sig++) {
            signals[sig] = new Signal(sig, "SIGRT" + sig);
        }
    }
}
